from transonic.config import backend_default

from transonic.justintime import modules_backends


def test_set_backend_for_this_module():
    from .for_test_set_backend import ts, func

    assert ts.backend.name == "python"
    assert func() == 0

    module_name = "transonic.backends.for_test_set_backend"

    assert module_name in modules_backends["python"]

    if backend_default == "python":
        return

    assert module_name not in modules_backends[backend_default]



import sys
import os
from time import sleep
from shutil import rmtree

import numpy as np
import pytest

from transonic.backends import backends
from transonic.compiler import scheduler, wait_for_all_extensions
from transonic.justintime import modules
from transonic import mpi
from transonic.util import can_import_accelerator
from transonic.config import backend_default

backend = backends[backend_default]
scheduler.nb_cpus = 2

module_name = "_transonic_testing.for_test_justintime"

str_relative_path = module_name.replace(".", os.path.sep)

path_jit = backend.jit.path_base
path_jit_classes = backend.jit.path_class

path_jit_dir = path_jit / str_relative_path
path_classes_dir = path_jit_classes / str_relative_path
path_classes_dir1 = path_jit / path_jit_classes.name / str_relative_path

if mpi.rank == 0:
    if path_jit.exists():
        rmtree(path_jit_dir, ignore_errors=True)
    if path_classes_dir.exists():
        rmtree(path_classes_dir)
    if path_classes_dir1.exists():
        rmtree(path_classes_dir1)

mpi.barrier()


@pytest.mark.skipif(backend_default == "numba", reason="Not supported by Numba")
def test_jit():
    from _transonic_testing.for_test_justintime import func1

    a = np.arange(2)
    b = [1, 2]

    for _ in range(2):
        func1(a, b)
        sleep(0.1)

    wait_for_all_extensions()
    func1(a, b)


def test_fib():
    from _transonic_testing.for_test_justintime import fib, use_fib

    fib2 = fib(2)
    result = use_fib()
    wait_for_all_extensions()
    assert fib2 == fib(2)
    assert result == use_fib()

    fib15 = fib(1.5)
    wait_for_all_extensions()
    assert fib15 == fib(1.5)


def test_jit_simple():

    from _transonic_testing.for_test_justintime import func_identity

    func_identity(1)

    if not can_import_accelerator():
        return

    mod = modules[module_name]
    cjit = mod.jit_functions["func_identity"]

    for _ in range(100):
        func_identity(1)
        sleep(0.1)
        if not cjit.compiling:
            sleep(0.1)
            func_identity(1)
            break

    del sys.modules[module_name]
    del modules[module_name]

    from _transonic_testing.for_test_justintime import func_identity

    func_identity(1)


def test_jit_array_bool():
    from _transonic_testing.for_test_justintime import func_identity

    a = np.zeros(2, dtype=bool)
    func_identity(a)
    wait_for_all_extensions()
    assert np.allclose(a, func_identity(a))


@pytest.mark.skipif(backend_default == "numba", reason="Not supported by Numba")
def test_jit_dict():
    from _transonic_testing.for_test_justintime import func_dict

    d = dict(a=1, b=2)
    func_dict(d)

    if not can_import_accelerator():
        return

    mod = modules[module_name]
    cjit = mod.jit_functions["func_dict"]

    d = dict(a=1, b=2)
    func_dict(d)

    wait_for_all_extensions()

    for _ in range(100):
        d = dict(a=1, b=2)
        func_dict(d)
        sleep(0.1)
        if not cjit.compiling:
            sleep(0.1)
            func_dict(d)
            break


def test_jit_method():
    from _transonic_testing.for_test_justintime import MyClass

    obj = MyClass()
    obj.check()

    if not can_import_accelerator():
        return

    obj = MyClass()
    obj.check()

    wait_for_all_extensions()

    obj.check()


@pytest.mark.skipif(backend_default == "numba", reason="Not supported by Numba")
def test_jit_method2():
    from _transonic_testing.for_test_justintime import MyClass2

    obj = MyClass2()
    obj.check()

    if not can_import_accelerator():
        return

    obj = MyClass2()
    obj.check()

    wait_for_all_extensions()

    obj.check()


def test_func0():
    from _transonic_testing.for_test_justintime import func0, func0_jitted

    func02 = func0(2.1)
    result = func0_jitted(2.1)
    wait_for_all_extensions()
    assert func02 == func0(2.1)
    assert result == func0(2.1)


# jitted function that uses a local function, a jitted local function
# and a jitted imported function (with a different name)
def test_main():
    from _transonic_testing.for_test_justintime import main

    main_res = main(2)
    wait_for_all_extensions()
    assert main_res == 5


def test_jit_imported():
    from _transonic_testing.for_test_justintime import (
        jitted_func_import,
        func_import,
    )

    result = jitted_func_import()
    wait_for_all_extensions()
    assert result == jitted_func_import() == func_import()



from transonic.analyses import get_types_from_transonic_signature


def assert_types_from_sig(signature, func_name, types):
    assert get_types_from_transonic_signature(signature, func_name) == types


def test_simple():
    assert_types_from_sig(
        "my_func(int, float64, )", "my_func", ["int", "float64"]
    )


def test_cython_style():
    assert_types_from_sig(
        "foo(int, int, int64[:, :], float64[:, :], float64[:, :], int64[:], int64[:, :, :], float64, float64, int, float64 )",
        "foo",
        [
            "int",
            "int",
            "int64[:, :]",
            "float64[:, :]",
            "float64[:, :]",
            "int64[:]",
            "int64[:, :, :]",
            "float64",
            "float64",
            "int",
            "float64",
        ],
    )


def test_c_style():
    assert_types_from_sig(
        "bar(int, int[] [] , int64[:, :], float64 )",
        "bar",
        ["int", "int[] []", "int64[:, :]", "float64"],
    )



from transonic.analyses.extast import parse
from transonic.analyses.util import print_dumped, print_unparsed

source = "def func(): return True"


def test_print_dumped():
    print_dumped(source)


def test_print_unparsed():
    node = parse(source)
    print_unparsed(node)



import numpy as np

from transonic import Array

from .typing import base_type_formatter


def compare(dtype, ndim, mem_layout, result):
    A = Array[dtype, ndim, mem_layout]
    assert A.format_as_backend_type(base_type_formatter) == result


def test_array():
    compare(int, "2d", "C", "int[:, :] order(C)")
    compare(int, "3d", "strided", "int[::, ::, ::]")
    compare(np.int32, "2d", "F", "int32[:, :] order(F)")



import shutil
from setuptools import Distribution
from pprint import pformat

import pytest

from transonic.config import backend_default

from transonic.dist import (
    detect_transonic_extensions,
    modification_date,
    make_backend_files,
    ParallelBuildExt,
    get_logger,
)

from transonic.mpi import nb_proc
from transonic.path_data_tests import path_data_tests
from transonic.util import can_import_accelerator


@pytest.mark.skipif(not path_data_tests.exists(), reason="no data tests")
@pytest.mark.skipif(nb_proc > 1, reason="No dist in MPI")
def test_detect_backend_extensions():

    shutil.rmtree(path_data_tests / f"__{backend_default}__", ignore_errors=True)

    names = [
        "assign_func_boost.py",
        "assign_func_jit.py",
        "block_fluidsim.py",
        "blocks_type_hints.py",
        "boosted_func_use_import.py",
        # "boosted_class_use_import.py",  # was forgotten...
        "class_blocks.py",
        "classic.py",
        # "class_rec_calls.py",
        # "methods.py",
        "mixed_classic_type_hint.py",
        # "no_arg.py",
        "type_hint_notemplate.py",
        "no_pythran_.py",
    ]

    make_backend_files((path_data_tests / name for name in names))
    ext_names = detect_transonic_extensions(path_data_tests)

    if can_import_accelerator():
        # -2 files (no_pythran.py and assign_fun_jit.py)
        number_not_transonized = 2

        if len(ext_names) != len(names) - number_not_transonized:
            print("ext_names:\n", pformat(sorted(ext_names)), sep="")
            print("names:\n", pformat(sorted(names)), sep="")
            raise RuntimeError

    shutil.rmtree(path_data_tests / f"__{backend_default}__", ignore_errors=True)


@pytest.mark.skipif(not path_data_tests.exists(), reason="no data tests")
@pytest.mark.skipif(nb_proc > 1, reason="No dist in MPI")
def test_modification_date():

    modification_date(path_data_tests / "no_pythran_.py")
    get_logger("bar")


@pytest.mark.skipif(not path_data_tests.exists(), reason="no data tests")
@pytest.mark.skipif(nb_proc > 1, reason="No dist in MPI")
def test_build_ext():
    dist = Distribution()
    build_ext = ParallelBuildExt(dist)

    build_ext.initialize_options()
    build_ext.parallel = 1
    build_ext.finalize_options()



import numpy as np

from transonic.typing import (
    Array,
    NDim,
    str2type,
    UnionMeta,
    List,
    ListMeta,
    Dict,
    DictMeta,
    Set,
    SetMeta,
    typeof,
    str2shape,
    MemLayout,
    Optional,
    const,
)

from transonic.backends.typing import base_type_formatter


def compare_array_types(A0, A1):

    assert A0.dtype == A1.dtype

    if len(A0.ndim.values) > 1:
        raise NotImplementedError

    if len(A1.ndim.values) > 1:
        raise NotImplementedError

    assert A0.ndim.values[0] == A1.ndim.values[0]


def test_NDim():
    N = NDim(1, 3)
    repr(N + 1)
    repr(N - 1)


def test_str2type_simple():
    assert str2type("int") == int
    assert str2type("float") == float
    assert str2type("uint32") == np.uint32


def test_str2type_arrays():
    A1 = Array[int, "1d"]
    compare_array_types(str2type("int[]"), A1)
    compare_array_types(str2type("int[:]"), A1)

    A2 = Array[np.uint32, "2d"]
    compare_array_types(str2type("uint32[:,:]"), A2)


def test_str2type_or():
    result = str2type("int or float")
    assert isinstance(result, UnionMeta)
    assert result.types == (int, float)

    result = str2type("int or float[]")
    assert isinstance(result, UnionMeta)
    compare_array_types(result.types[1], Array[float, "1d"])


def test_list():
    L = List[List[int]]
    repr(L)
    assert isinstance(L, ListMeta)
    assert L.format_as_backend_type(base_type_formatter) == "int list list"


def test_dict():
    D = Dict[str, int]
    repr(D)
    assert isinstance(D, DictMeta)
    assert D.format_as_backend_type(base_type_formatter) == "str: int dict"


def test_set():
    str2type("int set")
    S = Set["str"]
    S.get_template_parameters()
    repr(S)
    assert isinstance(S, SetMeta)
    assert S.format_as_backend_type(base_type_formatter) == "str set"


def test_tuple():
    T = str2type("(int, float[:, :])")
    T.get_template_parameters()
    assert repr(T) == 'Tuple[int, Array[float, "2d"]]'
    assert T.format_as_backend_type(base_type_formatter) == "(int, float64[:, :])"


def test_typeof_simple():
    assert typeof(1) is int
    assert typeof(1.0) is float
    assert typeof(1j) is complex
    assert typeof("foo") is str


def test_typeof_list():
    L = typeof([[1, 2], [3, 4]])
    assert isinstance(L, ListMeta)
    assert L.format_as_backend_type(base_type_formatter) == "int list list"


def test_typeof_dict():
    D = typeof({"a": 0, "b": 1})
    assert isinstance(D, DictMeta)
    assert D.format_as_backend_type(base_type_formatter) == "str: int dict"


def test_typeof_set():
    S = typeof({"a", "b"})
    assert isinstance(S, SetMeta)
    assert S.format_as_backend_type(base_type_formatter) == "str set"


def test_typeof_array():
    A = typeof(np.ones((2, 2)))
    compare_array_types(A, Array[np.float64, "2d"])


def test_typeof_np_scalar():
    T = typeof(np.ones(1)[0])
    assert T is np.float64


def test_shape():

    assert str2shape("[]") == (None,)
    assert str2shape("[:]") == (None,)
    assert str2shape("[][]") == (None,) * 2
    assert str2shape("[][ ]") == (None,) * 2

    assert str2shape("[:,:]") == (None,) * 2
    assert str2shape("[: ,:,:, ]") == (None,) * 3
    assert str2shape("[3 ,:,:]") == (3, None, None)
    assert str2shape("[ : , :,3]") == (None, None, 3)

    A = Array[int, "[: ,:, 3]"]
    assert A.shape == (None, None, 3)
    assert A.ndim.values[0] == 3
    assert repr(A) == 'Array[int, "[:,:,3]"]'

    assert (
        base_type_formatter.make_array_code(
            int, 2, (3, None), False, MemLayout.C_or_F, positive_indices=False
        )
        == "int[3, :]"
    )


def test_optional():
    assert repr(Optional[int]) == "Union[int, None]"


def test_const():
    A = str2type("int[]")
    B = const(A)
    assert A.format_as_backend_type(
        base_type_formatter
    ) == B.format_as_backend_type(base_type_formatter)
    assert repr(B) == 'const(Array[int, "1d"])'



import sys
import time
import os
from shutil import rmtree

import pytest

from transonic import util
from transonic.config import backend_default
from transonic.mpi import nb_proc
from transonic.path_data_tests import path_data_tests
from transonic.run import run


path_dir_out = path_data_tests / f"__{backend_default}__"
header_suffixes = {"pythran": ".pythran", "cython": ".pxd"}


@pytest.mark.skipif(not path_data_tests.exists(), reason="no data tests")
@pytest.mark.skipif(nb_proc > 1, reason="No commandline in MPI")
def test_create_pythran_files():

    if path_dir_out.exists():
        rmtree(path_dir_out)

    if os.name != "nt":
        sys.argv = f"transonic -nc {path_data_tests / '*.py'}".split()
        run()

    sys.argv = f"transonic -nc {path_data_tests}".split()
    run()

    paths = tuple(path_data_tests.glob("*.py*"))
    sys.argv = ["transonic", "-nc"] + [str(path) for path in paths]
    run()

    # At this point, we can compare the produced files with saved files.
    # For exterior files, we can't compare cause transonic changes their names
    no_compare = [
        "no_pythran_.py",
        "assign_func_jit.py",
        "exterior_import_boost.py",
        "exterior_import_boost_2.py",
    ]

    try:
        from ast import unparse
    except ImportError:
        dir_saved_files = "saved__backend__"
    else:
        dir_saved_files = "saved__backend__py3.9"

    for path in paths:
        if path.name in no_compare:
            continue

        __backend__path = path.parent / f"__{backend_default}__" / path.name
        assert __backend__path.exists()

        saved_path = (
            path.parent / dir_saved_files / backend_default / path.name
        )
        assert saved_path.exists()

        with open(__backend__path) as file:
            code = file.read()
        with open(saved_path) as file:
            saved_code = file.read()

        code = code.split("__transonic__ = ", 1)[0]
        saved_code = saved_code.split("__transonic__ = ", 1)[0]

        if backend_default in header_suffixes:
            suffix = header_suffixes[backend_default]

            header_path = __backend__path.with_suffix(suffix)
            assert header_path.exists()
            with open(header_path) as file:
                code += file.read()

            header_path = saved_path.with_suffix(suffix)
            assert header_path.exists()
            with open(header_path) as file:
                saved_code += file.read()

        # warning: it is a very strong requirement!
        assert code == saved_code, path


@pytest.mark.skipif(not path_data_tests.exists(), reason="no data tests")
@pytest.mark.skipif(nb_proc > 1, reason="No commandline in MPI")
def test_create_pythran_simple():

    sys.argv = "transonic --version".split()
    run()

    sys.argv = ["transonic"]
    run()


@pytest.mark.skipif(not path_data_tests.exists(), reason="no data tests")
@pytest.mark.skipif(nb_proc > 1, reason="No commandline in MPI")
def test_create_trans_classic():

    util.input = lambda: "y"

    if path_dir_out.exists():
        rmtree(path_dir_out)

    path_file = path_data_tests / "classic.py"
    sys.argv = f"transonic -nc {path_file}".split()
    run()

    print("after first build")
    run()

    time.sleep(0.02)
    path_file.touch()
    print("after touch")

    run()

    path_file_pythran = path_data_tests / f"__{backend_default}__/classic.py"
    path_file_pythran.unlink()

    print("after unlink")
    run()

    sys.argv = f"transonic -cc {path_file}".split()
    run()

    sys.argv = f"transonic {path_file}".split()
    run()

    sys.argv = f"transonic -nc {path_file_pythran}".split()
    run()


@pytest.mark.skipif(not path_data_tests.exists(), reason="no data tests")
@pytest.mark.skipif(nb_proc > 1, reason="No commandline in MPI")
def test_create_pythran_subpackages():
    path_file = path_data_tests / "subpackages.py"
    sys.argv = f"transonic -nc {path_file}".split()
    run()



from transonic import util
from transonic.util import query_yes_no, timeit, print_versions, timeit_verbose


def test_query_yes_no():

    util.input = lambda: "y"

    query_yes_no("test", default="y")
    query_yes_no("test", default="n")
    query_yes_no("test")

    util.input = lambda: ""
    query_yes_no("test", default="y")

    query_yes_no("test", force=True)


def test_timeit():
    a = 1
    b = 2
    timeit("a + b", total_duration=0.001, globals=locals())


def test_timeit_verbose():
    a = 1
    b = 2
    norm = timeit_verbose("a + b", total_duration=0.001, globals=locals())
    timeit_verbose("a + b", total_duration=0.001, globals=locals(), norm=norm)


def test_print_versions():
    print_versions()



import numpy as np

from transonic import Array, const
from transonic.backends import backends

backend = backends["cython"]
type_formatter = backend.type_formatter


def compare(result, dtype, ndim, memview, mem_layout=None, positive_indices=None):
    A = Array[dtype, ndim, memview, mem_layout, positive_indices]
    assert A.format_as_backend_type(type_formatter) == result


def test_memview():
    memview = "memview"
    compare("np.int_t[:, ::1]", int, "2d", memview, "C")
    compare("np.int_t[:, :, :]", int, "3d", memview, "strided")
    compare("np.int32_t[::1, :]", np.int32, "2d", memview, "F")


def test_array():
    memview = None
    compare('np.ndarray[np.int_t, ndim=2, mode="c"]', int, "2d", memview, "C")
    compare("np.ndarray[np.int_t, ndim=3]", int, "3d", memview, "strided")
    compare(
        'np.ndarray[np.int32_t, ndim=2, mode="f"]', np.int32, "2d", memview, "F"
    )
    compare(
        "np.ndarray[np.int_t, ndim=2, negative_indices=False]",
        int,
        "2d",
        memview,
        positive_indices="positive_indices",
    )


def test_const():
    A = Array[int, "2d"]
    assert "const " + A.format_as_backend_type(type_formatter) == const(
        A
    ).format_as_backend_type(type_formatter)



import ast

from .extast import parse, unparse


def test_ExtSlice():
    code = "foo[:, b]"
    code_back = unparse(parse(code)).strip()
    assert code_back == code

    code_back = unparse(ast.parse(code)).strip()
    assert code_back == code


def test_comment():
    code = "# a comment"
    code_back = unparse(parse(code)).strip()
    assert code_back == code




import importlib
from shutil import rmtree

from transonic import Transonic, mpi
from transonic.compiler import wait_for_all_extensions
from transonic.config import backend_default
from transonic.mpi import Path


def test_not_transonified():

    path_for_test = (
        Path(__file__).parent.parent / "_transonic_testing/for_test_init.py"
    )
    path_output = path_for_test.parent / f"__{backend_default}__"

    if path_output.exists() and mpi.rank == 0:
        rmtree(path_output)
    mpi.barrier()

    from _transonic_testing import for_test_init

    importlib.reload(for_test_init)

    from _transonic_testing.for_test_init import func, func1, check_class

    func(1, 3.14)
    func1(1.1, 2.2)
    check_class()


def test_use_pythran_false():
    Transonic(use_transonified=False)


def test_assign_boosted_func():
    from _transonic_testing.for_test_init import func0, func0_boosted

    func02 = func0(2, 6)
    result = func0_boosted(2, 6)
    wait_for_all_extensions()
    assert func02 == func0(2, 6)
    assert result == func0(2, 6)
